/************************************************************************/
/*									*/
/*  This is the ALTRO USB library    				        */
/*									*/
/*  16. Mar. 04  MAJO  created						*/
/*									*/
/*******C 2004 - The software with that certain something****************/

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include "altro.h"


/***********/
/* Globals */
/***********/
static int fd, altro_is_open = 0;
u_int traceLevel, packageId;


/**************************/
u_int ALTRO_Open(char *node)
/**************************/
{  
  u_int ret;
  
  if (altro_is_open) 
  {
    altro_is_open++;             /* keep track of multiple open calls */
    return(RCC_ERROR_RETURN(0, ALTRO_SUCCESS));
  }

  /* open the error package */
  ret = rcc_error_init((err_pid)P_ID_ALTRO, ALTRO_err_get);
  if (ret) 
  {
    DEBUG_TEXT(P_ID_ALTRO, 5 ,("ALTRO_Open: Failed to open error package\n"));
    return(RCC_ERROR_RETURN(0, ALTRO_ERROR_FAIL)); 
  }
  DEBUG_TEXT(P_ID_ALTRO, 20 ,("ALTRO_Open: error package opened\n"));

  DEBUG_TEXT(P_ID_ALTRO, 20 ,("ALTRO_Open: Opening %s\n", node));
  if ((fd = open(node, O_RDWR)) < 0) 
  {
    perror("open");
    DEBUG_TEXT(P_ID_ALTRO, 20 ,("ALTRO_Open: failed to open %s\n", node));
    return(RCC_ERROR_RETURN(0, ALTRO_FILE)); 
  }
  DEBUG_TEXT(P_ID_ALTRO, 20 ,("ALTRO_Open: %s is open\n", node));
  altro_is_open = 1;
  return(ALTRO_SUCCESS);
}


/*********************/
u_int ALTRO_Close(void)
/*********************/
{
  ALTROISOPEN;
  
  if (altro_is_open > 1)
    altro_is_open--;
  else 
  {  
    close(fd);
    altro_is_open = 0;
  }
  
  return(ALTRO_SUCCESS);
}


/*************************************/
u_int ALTRO_Send(altro_bulk_out_t *out)
/*************************************/
{
  int ret, e2;

  ALTROISOPEN;

  ret = ioctl(fd, SEND_DATA, out);
  e2 = errno;
  if (ret) 
  {
    DEBUG_TEXT(P_ID_ALTRO,5,("ALTRO_Send: Error from ioctl, ret = 0x%08x: %s\n", -ret, strerror(-ret)));
    DEBUG_TEXT(P_ID_ALTRO,5,("ALTRO_Send: Error from ioctl, e2 = 0x%08x: %s\n", e2, strerror(errno)));
    return(RCC_ERROR_RETURN(0, -ret));
  }

  return(ALTRO_SUCCESS);
}


/**********************************/
u_int ALTRO_Get(altro_bulk_in_t *in)
/**********************************/
{
  int ret, e2;

  if (!altro_is_open) 
    return(RCC_ERROR_RETURN(0, ALTRO_NOTOPEN));

  ret = ioctl(fd, GET_DATA, in);
  e2 = errno;
  if (ret) 
  {
    DEBUG_TEXT(P_ID_ALTRO,5,("ALTRO_Get: Error from ioctl, ret = 0x%08x\n", ret));
    DEBUG_TEXT(P_ID_ALTRO,5,("ALTRO_Get: Error from ioctl, e2 = 0x%08x\n", e2));
   return(RCC_ERROR_RETURN(0, -ret));
  }

  return(ALTRO_SUCCESS);
}


/****************************************/
u_int ALTRO_Control(altro_control_t *ctrl)
/****************************************/
{
  int ret;

  if (!altro_is_open) 
    return(RCC_ERROR_RETURN(0, ALTRO_NOTOPEN));

  ret = ioctl(fd, CONTROL, ctrl);
  if (ret) 
  {
    DEBUG_TEXT(P_ID_ALTRO, 5 , ("ALTRO_Control: Error from ioctl, errno = 0x%08x\n", errno));
    return(RCC_ERROR_RETURN(0, -ret));
  }
  
  return(ALTRO_SUCCESS);
}


/*****************************************************************/
unsigned int ALTRO_err_get(err_pack err, err_str pid, err_str code)
/*****************************************************************/
{ 
  strcpy(pid, P_ID_ALTRO_STR);

  switch (RCC_ERROR_MINOR(err)) 
  {  
    case ALTRO_SUCCESS:  strcpy(code, ALTRO_SUCCESS_STR);   break;
    case ALTRO_FILE:     strcpy(code, ALTRO_FILE_STR);      break;
    case ALTRO_EFAULT:   strcpy(code, ALTRO_EFAULT_STR);    break;
    default:             strcpy(code, ALTRO_NO_CODE_STR);     return(RCC_ERROR_RETURN(0, ALTRO_NO_CODE)); break;
  }
  return(RCC_ERROR_RETURN(0, ALTRO_SUCCESS));
}
