/************************************************************************/
/*									*/
/*  This is the FEC library 	   				        */
/*									*/
/*  25. Jul. 05  L. Musa         					*/
/*									*/
/************************************************************************/

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <fcntl.h>
#include <math.h>
#include <tools/get_input.h>
#include <altro/altro.h>
#include <tools/rcc_time_stamp.h>
#include "u2f.h"
#include "fec.h"


/*****************************************************************************************/
u_int Write_ALTRO_Reg(u_int debug, u_int ch_addr, u_int reg_addr, u_int reg_data) 
/*****************************************************************************************/
{
  u_int ret;
  u_int memdata[256];


  /* Preparing the Instruction Memory data */
  if(debug)
    printf("Preparing the Instruction Memory data\n");
	
  memdata[0] = 0x00600000 + (ch_addr << 5) + reg_addr;	
  memdata[1] = 0x00700000 + reg_data;
  memdata[2] = 0x00390000; 
 
  if(debug)
    printf("Memory Data Block:\n %x\n %x\n %x\n", memdata[0], memdata[1],memdata[2]);
     
  /* Writing the Instruction Memory */
  if(debug)
    printf("Writing into the Instruction Memory\n");
  ret = U2F_IMEM_Write(3, 0, memdata);
  if (ret)
    rcc_error_print(stdout, ret);
     
  /* Execute the Instruction Memory */
  if(debug)
    printf("Executing the Instruction Memory\n");
  ret = U2F_Exec_Command(C_EXEC, 0);
  if (ret)
    rcc_error_print(stdout, ret);

  return ret;
}


/*****************************************************************************************/
u_int Read_ALTRO_Reg(u_int debug, u_int ch_addr, u_int reg_addr, u_int *reg_data) 
/*****************************************************************************************/
{
  u_int ret;
  u_int osize;
  u_int memdata[256];


  /* Preparing the Instruction Memory data */
  if(debug)
    printf("Preparing the Instruction Memory data\n");
	
  memdata[0] = 0x00500000 + (ch_addr << 5) + reg_addr;	
  memdata[1] = 0x00390000; 
 
  if(debug)
    printf("Memory Data Block:\n %x\n %x\n", memdata[0], memdata[1]);
     
  /* Writing the Instruction Memory */
  if(debug)
    printf("Writing into the Instruction Memory\n");
  ret = U2F_IMEM_Write(2, 0, memdata);
  if (ret)
    rcc_error_print(stdout, ret);
     
  /* Execute the Instruction Memory */
  if(debug)
    printf("Executing the Instruction Memory\n");
  ret = U2F_Exec_Command(C_EXEC, 0);
  if (ret)
    rcc_error_print(stdout, ret);

  /* Reading the Result memory */
  if(debug)
    printf("Reading the Result Memory\n");  
  ret = U2F_RMEM_Read(1, &osize, 0, reg_data);  
  if (ret)
    rcc_error_print(stdout, ret);
  if(debug)
    printf("Result Memory: %x\n", *reg_data); 

  return ret;
}

/*****************************************************************************************/
u_int Write_ALTRO_PMEM(u_int debug, u_int bcast, u_int ch_addr, u_int nwords, char *fname) 
/*****************************************************************************************/
{
  u_int ret;
  u_int loop;
  u_int memdata[1024];
  u_int fdata[1024];

  /* Reading the Pedestal Memory Pattern from file */
  if(debug)
    printf("Reading the Pedestal Memory Pattern from file:\n%s\n", fname);
  ret = U2F_File_Read(fname, nwords, fdata);
  if (ret)
  rcc_error_print(stdout, ret); 
    
  /* Preparing the Pattern Memory Block */
  if(debug)
    printf("Preparing the Pattern Memory Block\n");
  for(loop=0; loop < nwords; loop++)
  {
    memdata[loop] = fdata[loop];
    if(debug)
      printf("Data word %d = 0x%08x\n", loop, memdata[loop]);
  }
  ret = U2F_PMEM_Write(nwords, 0, memdata);
  if (ret)
    rcc_error_print(stdout, ret);


  /* Preparing the Instruction Memory data */
  if(debug)
    printf("Preparing the Instruction Memory data\n");
  memdata[0] = 0x00380000 + (bcast << 12) + ch_addr;	
  memdata[1] = 0x00390000; 
  if(debug)
    printf("Memory Data Block:\n %x\n %x\n", memdata[0], memdata[1]);
     
  /* Writing the Instruction Memory */
  if(debug)
    printf("Writing into the Instruction Memory\n");
  ret = U2F_IMEM_Write(2, 0, memdata);
  if (ret)
    rcc_error_print(stdout, ret);
     
  /* Execute the Instruction Memory */
  if(debug)
    printf("Executing the Instruction Memory\n");
  ret = U2F_Exec_Command(C_EXEC, 0);
  if (ret)
    rcc_error_print(stdout, ret);

  return ret;
}


/*****************************************************************************************/
u_int Read_ALTRO_PMEM(u_int debug, u_int ch_addr, u_int nwords, char fname[]) 
/*****************************************************************************************/
{
  u_int ret;
  u_int loop;
  u_int osize;
  u_int reg_data[256];
  u_int memdata[1024];
  u_int fdata[1024];


  for(loop = 0; loop < nwords; loop++)
  {
    /* Preparing the Instruction Memory data */
    if(debug)
      printf("Preparing the Instruction Memory data\n");
    memdata[0] = 0x0060000d + (ch_addr << 5);	
    memdata[1] = 0x00700000 + loop;
    memdata[2] = 0x00500007 + (ch_addr << 5);
    memdata[3] = 0x00390000; 
    if(debug)
      printf("Memory Data Block:\n %x\n %x\n %x\n %x\n", memdata[0], memdata[1], 
              memdata[2], memdata[3]);    
 
    /* Writing the Instruction Memory */
    if(debug)
      printf("Writing into the Instruction Memory\n");
    ret = U2F_IMEM_Write(4, 0, memdata);
    if (ret)
      rcc_error_print(stdout, ret);
    
    /* Execute the Instruction Memory */
    if(debug)
      printf("Executing the Instruction Memory\n");
    ret = U2F_Exec_Command(C_EXEC, 0); 
    if (ret)
      rcc_error_print(stdout, ret);

    /* Reading the Result memory */
    if(debug)
      printf("Reading the Result Memory\n");  
    ret = U2F_RMEM_Read(1, &osize, 0, reg_data); 
    if (ret)
      rcc_error_print(stdout, ret);
    if(debug)
      printf("Result Memory: %x\n", *reg_data);
      
    /* copying the content of the result memory into fdata */
   
    fdata[loop] = *reg_data;
    if(debug)
      printf("Pedestal Data:\n %x\n", fdata[loop]);
      
  }
  
  /* Writing the pedestal pattern into file fname */   
    if(debug)
      printf("Writing Pedestal Data into file\n");
    ret = U2F_File_Write(fname, nwords, fdata);
    
  return ret;
}

  

/*********************************************************************/
u_int FEC_ON_RESET(u_int debug)
/*********************************************************************/
{
  u_int ret;

  /* Open the rcc_time_stamp library */
  ret = ts_open(1, TS_DUMMY);
  if (ret)
    rcc_error_print(stdout, ret);

  /* Switch on FEC with Address = 0  */
  if(debug)
    printf("FEC Switched ON\n");
  ret = U2F_Reg_Write(O_ACTFEC, 0x1);
  if (ret)
    rcc_error_print(stdout, ret);
    
  /* delay 1 sec */
  ts_delay(1000000);

  /* Reset FEC */
  if(debug)
    printf("FEC Has been reset\n");
  ret = U2F_Exec_Command(C_FECRST, 0x0);
  if (ret)
    rcc_error_print(stdout, ret);
    
  /* delay 10 us */
  ts_delay(10);
  
  return ret;
}




/************************************************************************************/
/************************************************************************************/
/*				FEC_Event_Readout				    */
/************************************************************************************/
/************************************************************************************/

u_int FEC_Event_Readout(u_int debug, u_int trg_mode, u_int rdo_mode, u_int nevents, u_int nsamples, char fname[]) 
{
  u_int ret;
  u_int loop, nev;
  u_int chnum =0, tick=0, point=0;
  u_int rdata=0;
  u_int osize=0;
  u_int rsize=0;
  double int_nwords=0, float_nwords =0.0, frac_nwords=0.0;
  u_int memdata[200000];
  FILE *file;
  
  /*************************************************************************/
  /*				FEC INITIALIZATION			   */
  /*************************************************************************/

  /* Preparing the Instruction Memory data */
  if(debug)
    printf("\nPreparing the Instruction Memory Data\n");
	
  memdata[0] = 0x0064000a;		
  memdata[1] = 0x00700000 + nsamples;
  if(rdo_mode)
  {
    memdata[2] = 0x0064000b; 
    memdata[3] = 0x0070000a;
  }
  else
  {
    memdata[2] = 0x0064000b; 
    memdata[3] = 0x00700000;
  }  
  memdata[4] = 0x00390000; 
 
  if(debug)
  {
    printf("\nMemory Data Block:\n");
    for(loop = 0; loop < 4; loop++)
    {
      printf("Memory Data[%d] =\t%d\n", loop, memdata[loop]);
    }
  }
     
  /* Writing the Instruction Memory */
  if(debug)
    printf("\nWriting into the Instruction Memory\n");
  ret = U2F_IMEM_Write(5, 0, memdata);
  if (ret)
    rcc_error_print(stdout, ret);
     
  /* Execute the Instruction Memory */
  if(debug)
    printf("\nExecuting the Instruction Memory\n");
  ret = U2F_Exec_Command(C_EXEC, 0);
  if (ret)
    rcc_error_print(stdout, ret);

  /* delay 1ms */
  ts_delay(1000);


  /*************************************************************************/
  /*			INITIALIZATION the U2F				   */
  /*************************************************************************/
  
  /* Initializing the Active Channel List */
  if(debug)
    printf("\nInitializing the Channel Readout List\n");
  memdata[0] = 0xffff;
  memdata[1] = 0xffff;
  memdata[2] = 0xffff;
  memdata[3] = 0xffff;
  memdata[4] = 0xffff;
  memdata[5] = 0xffff;
  memdata[6] = 0xffff;
  memdata[7] = 0xffff;    
  
  ret = U2F_ACL_Write(8, 0, memdata);
  if (ret)
    rcc_error_print(stdout, ret);

  /* Configuring the Trigger Configuration Register */
  if(debug)
    printf("\nConfiguring the Trigger Configuration Register\n");
  ret = U2F_Reg_Write(O_TRCFG, 0x50800);
  if (ret)
    rcc_error_print(stdout, ret);

  /*************************************************************************/
  /*			Sending a TRIGGER				   */
  /*************************************************************************/

  /*** Opening the Data File ***/
  file = fopen(fname,"w"); 

for(nev = 0; nev < nevents; nev++)
{ 
  point =0;
  chnum =0;  
  /* Send a software trigger */
  if(trg_mode)
  {
    ret = U2F_Exec_Command(C_SWTRG, 0);
    if (ret)
      rcc_error_print(stdout, ret);
    if(debug)
      printf("\n\nSOFTWARE TRIGGER ISSUED\n\n");
  }

  /*** Waiting for an external Trigger ***/
  /***  Read EVWORD Register  ***/
  else
  {

    ret = U2F_Reg_Write(O_RDOMOD, 0x2); 
    if (ret)
      rcc_error_print(stdout, ret);

    ret = U2F_Reg_Read(O_RDOMOD, &rdata); 
    if (ret)
      rcc_error_print(stdout, ret);
    if(debug)
      printf("External Trigger Enabled = %x\n", rdata);

 
    ret = U2F_Reg_Read(O_EVWORD, &rdata); 
    if (ret)
      rcc_error_print(stdout, ret);
    if(debug)
      printf("EVWORD = %x\n", rdata);

    while(!(rdata & 0x00000010))
    {
      tick++;

      /***  Read EVWORD Register  ***/
      ret = U2F_Reg_Read(O_EVWORD, &rdata); 
      if (ret)
        rcc_error_print(stdout, ret);
      if(tick == 5000)
      {
        if(debug)
          printf("Waiting for external trigger, point = %d, tick=%d\n", point, tick);
	point++;
	tick = 0;
      } 
      if(point == 10)
      {
        printf("point = %d, tick=%d\n", point,tick);

        ret = U2F_Reg_Write(O_RDOMOD, 0x0); 
        if (ret)
          rcc_error_print(stdout, ret);

        ret = U2F_Reg_Read(O_RDOMOD, &rdata); 
        if (ret)
          rcc_error_print(stdout, ret);
        if(debug)
          printf("External Trigger Disabled = %x\n", rdata);

        return(0);  		
      }
    }
  }


  /*************************************************************************/
  /*			Perform the Readout				   */
  /*************************************************************************/
 
  /***  Read EVWORD Register  ***/
  ret = U2F_Reg_Read(O_EVWORD, &rdata); 
  if (ret)
    rcc_error_print(stdout, ret);
  if(debug)
    printf("EVWORD = %x\n", rdata);

  if((rdata & 0x00000010) && (rdata & 0x00000008)) 
  {
    if(debug)
    {
      printf("STARTING EVENT READOUT\n");  	
      if(rdata & 0x00000004)
  	printf("EVENT READOUT WAS CONCLUDED WITHOUT DATA TRANSFER\n");
    }

    /*** Calculating Data Block Size ***/  
    float_nwords =  (nsamples + 2) / 4.0;
    frac_nwords = modf(float_nwords, &int_nwords);
    if(frac_nwords)
      rsize = int_nwords + 2;
    else
      rsize = int_nwords +1; 
    if(debug)
      printf("\nData Block Size = %d\n", rsize);

    while(!(rdata & 0x00000004))
    {
      /*******************************/
      /***  Read EVWORD Register  ***/
      /******************************/
      ret = U2F_Reg_Read(O_EVWORD, &rdata); 
      if (ret)
        rcc_error_print(stdout, ret);

      /*******************************/
      /***  Reading Data  	   ***/
      /*******************************/

      /*** Reading DMEM1  ***/

      if (rdata & 0x00000002)
      {
	if(debug)
	  printf("\n\nReading Channel #%d from DMEM1\n", chnum);

	ret = U2F_DM1_Read(rsize, &osize, 0, memdata);
	if (ret)
	  rcc_error_print(stdout, ret);

	for (loop = 0; loop < osize; loop++) 
	{
	  fprintf(file, "%d\n", memdata[loop * 4]);
	  fprintf(file, "%d\n", memdata[loop * 4 +1]);
	  fprintf(file, "%d\n", memdata[loop * 4 +2]);
	  fprintf(file, "%d\n", memdata[loop * 4 +3]);	
	}

	/* Read EVWORD */ 
	ret = U2F_Reg_Read(O_EVWORD, &rdata); 
	if (ret)
	  rcc_error_print(stdout, ret);
	if(debug)
	  printf("EVWORD = %x\n", rdata);

	/***  Reset DMEM1 flag    ***/
	ret = U2F_Exec_Command(C_RS_DMEM1, 0);
	if (ret)
	  rcc_error_print(stdout, ret);

	/* Read EVWORD */ 
	ret = U2F_Reg_Read(O_EVWORD, &rdata); 
	if (ret)
	  rcc_error_print(stdout, ret);
	if(debug)
	  printf("EVWORD = %x\n", rdata);
      }

      /***  Reading DMEM2  ***/
      if (rdata & 0x00000001)
      {
	chnum++;
	if(debug)	
	  printf("\n\nReading Channel #%d from DMEM2\n", chnum);

	ret = U2F_DM2_Read(rsize, &osize, 0, memdata); 
	if (ret)
	  rcc_error_print(stdout, ret);

	for (loop = 0; loop < osize; loop++) 
	{
	  fprintf(file, "%d\n", memdata[loop * 4]);
	  fprintf(file, "%d\n", memdata[loop * 4 +1]);
	  fprintf(file, "%d\n", memdata[loop * 4 +2]);
	  fprintf(file, "%d\n", memdata[loop * 4 +3]);	
	}

	/* Read EVWORD */ 
	ret = U2F_Reg_Read(O_EVWORD, &rdata); 
	if (ret)
	  rcc_error_print(stdout, ret);
	if(debug)
	  printf("EVWORD = %x\n", rdata);

	/***  Reset DMEM2 flag    ***/
	ret = U2F_Exec_Command(C_RS_DMEM2, 0);
	if (ret)
	  rcc_error_print(stdout, ret);

	/* Read EVWORD */ 
	ret = U2F_Reg_Read(O_EVWORD, &rdata); 
	if (ret)
	  rcc_error_print(stdout, ret);
	if(debug)
	  printf("EVWORD = %x\n", rdata);
      }
      chnum++;
    }

    /*** Reset EVWORD ***/
    ret = U2F_Exec_Command(C_TRG_CLR, 0);
    if (ret)
      rcc_error_print(stdout, ret);

    /*** Read EVWORD ***/ 
    ret = U2F_Reg_Read(O_EVWORD, &rdata); 
    if (ret)
      rcc_error_print(stdout, ret);
    if(debug)
      printf("EVWORD = %x\n", rdata);
  }
}
  fclose(file); 

  ret = U2F_Reg_Write(O_RDOMOD, 0x0); 
  if (ret)
    rcc_error_print(stdout, ret);

  ret = U2F_Reg_Read(O_RDOMOD, &rdata); 
  if (ret)
    rcc_error_print(stdout, ret);
  if(debug)
    printf("External Trigger Disabled = %x\n", rdata);
    
  return ret;
}


/************************************************************************************/
/************************************************************************************/
/*				FEC_BC_Reg_Write				    */
/************************************************************************************/
/************************************************************************************/

u_int FEC_BC_Reg_Write(u_int debug, u_int fec_addr, u_int reg_addr, u_int reg_data) 
{
  u_int ret;
  u_int memdata[256];


  /* Preparing the Instruction Memory data */
  if(debug)
    printf("Preparing the Instruction Memory data\n");
	
  memdata[0] = 0x00620000 + ((fec_addr & 0x1f) << 12) + (reg_addr & 0x1f );	
  memdata[1] = 0x00700000 + (reg_data & 0xfffff);
  memdata[2] = 0x00390000; 
 
  if(debug)
    printf("Memory Data Block:\n %x\n %x\n %x\n", memdata[0], memdata[1],memdata[2]);
     
  /* Writing the Instruction Memory */
  if(debug)
    printf("Writing into the Instruction Memory\n");
  ret = U2F_IMEM_Write(3, 0, memdata);
  if (ret)
    rcc_error_print(stdout, ret);
     
  /* Execute the Instruction Memory */
  if(debug)
    printf("Executing the Instruction Memory\n");
  ret = U2F_Exec_Command(C_EXEC, 0);
  if (ret)
    rcc_error_print(stdout, ret);

  return ret;
}


/************************************************************************************/
/************************************************************************************/
/*				FEC_BC_Reg_Read					    */
/************************************************************************************/
/************************************************************************************/

u_int FEC_BC_Reg_Read(u_int debug, u_int fec_addr, u_int reg_addr, u_int *reg_data) 
{
  u_int ret;
  u_int osize;
  u_int memdata[256];
  
  u_int csr3_data = 0;
  u_int conv_done = 0;
  u_int conv_cycle = 0;


  /* Reading Registers that require  to be updated by running the mADC */
  if((reg_addr == 0x06) || (reg_addr == 0x07) || (reg_addr == 0x08) || (reg_addr == 0x09) || (reg_addr == 0x0a))
  {

    ret = FEC_BC_Reg_Write(debug, fec_addr, 0x1b, 0);
    
    while(!conv_done)
    {
      /* delay 1ms */
      ts_delay(1000);

      ret = FEC_BC_Reg_Read(debug, fec_addr, 0x14, &csr3_data);  
      conv_done = csr3_data & 0x8000;
      conv_cycle++;
      
      if(conv_cycle == 10)
        exit(0);
    }
    
    /* Preparing the Instruction Memory data */
    if(debug)
      printf("Preparing the Instruction Memory data\n");
	
    memdata[0] = 0x00520000 + ((fec_addr & 0x1f) << 12) + (reg_addr & 0x1f);	
    memdata[1] = 0x00390000; 
 
    if(debug)
      printf("Memory Data Block:\n %x\n %x\n", memdata[0], memdata[1]);
     
    /* Writing the Instruction Memory */
    if(debug)
      printf("Writing into the Instruction Memory\n");
    ret = U2F_IMEM_Write(2, 0, memdata);
    if (ret)
      rcc_error_print(stdout, ret);
     
    /* Execute the Instruction Memory */
    if(debug)
      printf("Executing the Instruction Memory\n");
    ret = U2F_Exec_Command(C_EXEC, 0);
    if (ret)
      rcc_error_print(stdout, ret);

    /* Reading the Result memory */
    if(debug)
      printf("Reading the Result Memory\n");  
    ret = U2F_RMEM_Read(1, &osize, 0, reg_data);  
    if (ret)
      rcc_error_print(stdout, ret);
    if(debug)
      printf("Result Memory: %x\n", *reg_data); 
  }    
 
  /*  Reading any other BC Register  */
  else 
  {
    /* Preparing the Instruction Memory data */
    if(debug)
      printf("Preparing the Instruction Memory data\n");
	
    memdata[0] = 0x00520000 + ((fec_addr & 0x1f) << 12) + (reg_addr & 0x1f);	
    memdata[1] = 0x00390000; 
 
    if(debug)
      printf("Memory Data Block:\n %x\n %x\n", memdata[0], memdata[1]);
     
    /* Writing the Instruction Memory */
    if(debug)
      printf("Writing into the Instruction Memory\n");
    ret = U2F_IMEM_Write(2, 0, memdata);
    if (ret)
      rcc_error_print(stdout, ret);
     
    /* Execute the Instruction Memory */
    if(debug)
      printf("Executing the Instruction Memory\n");
    ret = U2F_Exec_Command(C_EXEC, 0);
    if (ret)
      rcc_error_print(stdout, ret);

    /* Reading the Result memory */
    if(debug)
      printf("Reading the Result Memory\n");  
    ret = U2F_RMEM_Read(1, &osize, 0, reg_data);  
    if (ret)
      rcc_error_print(stdout, ret);
    if(debug)
      printf("Result Memory: %x\n", *reg_data); 
  }      
  return ret;
}

/************************************************************************************/
/************************************************************************************/
/*				FEC_BC_Mon					    */
/************************************************************************************/
/************************************************************************************/
u_int FEC_BC_Mon(u_int debug, u_int fec_addr, float *temp, float *ana_v,  float *ana_i, float *dig_v, float *dig_i) 
{
  u_int ret;
  u_int reg_temp = 0;
  u_int reg_ana_v = 0;
  u_int reg_ana_i = 0;
  u_int reg_dig_v = 0;
  u_int reg_dig_i = 0;

  ret = FEC_BC_Reg_Read(debug, fec_addr, 0x06, &reg_temp);
  ret = FEC_BC_Reg_Read(debug, fec_addr, 0x07, &reg_ana_v);
  ret = FEC_BC_Reg_Read(debug, fec_addr, 0x08, &reg_ana_i);
  ret = FEC_BC_Reg_Read(debug, fec_addr, 0x09, &reg_dig_v);
  ret = FEC_BC_Reg_Read(debug, fec_addr, 0x0a, &reg_dig_i);

  *temp  = reg_temp / 4.0;
  *ana_v = reg_ana_v * 4.43;
  *ana_i = reg_ana_i * 17.0;
  *dig_v = reg_dig_v * 4.43;
  *dig_i = reg_dig_i * 30.0;

   
  if(debug)
  {
  	printf("\nFEC Monitoring ADC:\n");
  	printf("Temperature =\t%.1fC\n", *temp);
  	printf("Analogue Voltage =\t%.0f mV\n", *ana_v);
  	printf("Analogue Current =\t%.0f mA\n", *ana_i);
  	printf("Digital Voltage =\t%.0f mV\n", *dig_v);
  	printf("Digital Current =\t%.0f mA\n", *dig_i);
  }
  return ret;
}  


/*************************************************************************************/
/************************************************************************************/
/*				FEC_FM_Write					    */
/************************************************************************************/
/************************************************************************************/
u_int FEC_FM_Write(u_int debug, u_int fec_addr, u_int nwords, char fname[])
{
  u_int ret;
  u_int loop;
  u_int fm_address, fm_data; 
  u_int fdata[128];
  

  /*** Reading file fname ***/
  ret = U2F_File_Read(fname, nwords, fdata);
  if (ret)
    rcc_error_print(stdout, ret);  

  /* Writing in the FM */   
  if(debug)
  {
    printf("\n Writing %d words in the FM of FEC %d\n", nwords, fec_addr);
    for(loop = 0; loop < nwords; loop++)
    {
      fm_address = (fdata[loop] >> 8) & 0xff;
      fm_data = fdata[loop] & 0xff;
      printf("Address = %d\t\t Data = %d\n", fm_address, fm_data);
    }       
  }    
  
  for(loop = 0; loop < nwords; loop++)
  {
    /* Writing FM Address and Data in FMIREG */
    ret = U2F_Reg_Write(O_FMIREG, fdata[loop]);
    if (ret)
    rcc_error_print(stdout, ret);
    
    /* Execute the WRFM Command */
    ret = U2F_Exec_Command(C_WRFM, 0);
    if (ret)
      rcc_error_print(stdout, ret);
    
    /* delay 5 ms */
    ts_delay(5000);
   }
  
  if(debug)
    printf("\n FEC's Flash Memory Written\n");
    
  return ret;
}


/*************************************************************************************/
/************************************************************************************/
/*				FEC_FM_Read					    */
/************************************************************************************/
/************************************************************************************/
u_int FEC_FM_Read(u_int debug, u_int fec_addr, u_int nwords, char fname[])
{
  u_int ret;
  u_int loop;
  u_int fm_address;
  u_int rdata = 0; 
  u_int fdata[128];
  

  /* Reading in the FM */   
  if(debug)
    printf("\n Reading %d words in the FM of FEC %d\n", nwords, fec_addr);
  
  for(loop = 0; loop < nwords; loop++)
  {
    /* Writing FM Address in FMIREG */
    fm_address = (loop << 8) & 0xffff;
    ret = U2F_Reg_Write(O_FMIREG, fm_address);
    if (ret)
    rcc_error_print(stdout, ret);
    
    /* Execute the RDFM Command */
    ret = U2F_Exec_Command(C_RDFM, 0);
    if (ret)
      rcc_error_print(stdout, ret);
    
    /* delay 5 ms */
    ts_delay(5000);

    /* Reading FMOREG */
    ret = U2F_Reg_Read(O_FMOREG, &rdata);
    if (ret)
      rcc_error_print(stdout, ret);

    fdata[loop] = (((loop << 8) & 0xff00) | rdata) & 0xffff ;
    
    if(debug)
      printf("Address = %d\t\t Data = %d\n", loop, rdata & 0xff);
  }
    
  /* Writing FM data in file fname */
  if(debug)
    printf("Writing Flash Memory Data into file\n");
  ret = U2F_File_Write(fname, nwords, fdata);

  return ret;
}



/*********************************************************************************************************************************/
/*********************************************************************************************************************************/
/*				                        FEC_Event_Readout_2                                                      */
/*********************************************************************************************************************************/
/*********************************************************************************************************************************/

u_int FEC_Event_Readout_2(u_int debug, u_int fec_addr, u_int trg_mode, u_int rdo_mode, u_int nevents, u_int nsamples, char fname[]) 
{
  u_int ret;
  u_int loop, nev, index=0;
  u_int chnum =0, tick=0, point=0;
  u_int rdata=0;
  u_int osize=0;
  u_int rsize=0;
  double int_nwords=0, float_nwords =0.0, frac_nwords=0.0;
  u_int memdata[200000];
  FILE *file;
  
  /*************************************************************************/
  /*				FEC INITIALIZATION			   */
  /*************************************************************************/

  /* Preparing the Instruction Memory data */
  if(debug)
    printf("\nPreparing the Instruction Memory Data\n");
	
  memdata[0] = 0x0064000a;		
  memdata[1] = 0x00700000 + nsamples;
  memdata[2] = 0x0064000c;
  memdata[3] = 0x0070000d;
  if(rdo_mode)
  {
    memdata[4] = 0x0064000b; 
    memdata[5] = 0x0070000a;
  }
  else
  {
    memdata[4] = 0x0064000b; 
    memdata[5] = 0x00700000;
  }  
  memdata[6] = 0x00390000; 
 
  if(debug)
  {
    printf("\nMemory Data Block:\n");
    for(loop = 0; loop < 6; loop++)
    {
      printf("Memory Data[%d] =\t%x\n", loop, memdata[loop]);
    }
  }
     
  /* Writing the Instruction Memory */
  if(debug)
    printf("\nWriting into the Instruction Memory\n");
  ret = U2F_IMEM_Write(5, 0, memdata);
  if (ret)
    rcc_error_print(stdout, ret);
         
  /* Execute the Instruction Memory */
  if(debug)
    printf("\nExecuting the Instruction Memory\n");
  ret = U2F_Exec_Command(C_EXEC, 0);
  if (ret)
    rcc_error_print(stdout, ret);

  /* delay 1ms */
  ts_delay(1000);


  /*************************************************************************/
  /*			INITIALIZING the U2F				   */
  /*************************************************************************/
  
  /* Initializing the Readout List */
  if(debug)
    printf("\nInitializing the Channel Readout List\n");
  for(loop =0; loop < 128; loop++)
    memdata[loop] = 0x0;
  ret = U2F_ACL_Write(128, 0, memdata);
  if (ret)
    rcc_error_print(stdout, ret);
    
  memdata[0] = 0xffff;
  memdata[1] = 0xffff;
  memdata[2] = 0xffff;
  memdata[3] = 0xffff;
  memdata[4] = 0xffff;
  memdata[5] = 0xffff;
  memdata[6] = 0xffff;
  memdata[7] = 0xffff;    
  
  index = fec_addr * 8;
  ret = U2F_ACL_Write(8, index, memdata);
  if (ret)
    rcc_error_print(stdout, ret);

  /* Configuring the Trigger Configuration Register */
  if(debug)
    printf("\nConfiguring the Trigger Configuration Register\n");
  ret = U2F_Reg_Write(O_TRCFG, 0x50800);
  if (ret)
    rcc_error_print(stdout, ret);

  /*************************************************************************/
  /*			Sending a TRIGGER				   */
  /*************************************************************************/

  /*** Opening the Data File ***/
  file = fopen(fname,"w"); 

for(nev = 0; nev < nevents; nev++)
{ 
  point =0;
  chnum =0;  
  /* Send a software trigger */
  if(trg_mode)
  {
    ret = U2F_Exec_Command(C_SWTRG, 0);
    if (ret)
      rcc_error_print(stdout, ret);
    if(debug)
      printf("\n\nSOFTWARE TRIGGER ISSUED\n\n");
  }

  /*** Waiting for an external Trigger ***/
  /***  Read EVWORD Register  ***/
  else
  {

    ret = U2F_Reg_Write(O_RDOMOD, 0x2); 
    if (ret)
      rcc_error_print(stdout, ret);

    ret = U2F_Reg_Read(O_RDOMOD, &rdata); 
    if (ret)
      rcc_error_print(stdout, ret);
    if(debug)
      printf("External Trigger Enabled = %x\n", rdata);

 
    ret = U2F_Reg_Read(O_EVWORD, &rdata); 
    if (ret)
      rcc_error_print(stdout, ret);
    if(debug)
      printf("EVWORD = %x\n", rdata);

    while(!(rdata & 0x00000010))
    {
      tick++;

      /***  Read EVWORD Register  ***/
      ret = U2F_Reg_Read(O_EVWORD, &rdata); 
      if (ret)
        rcc_error_print(stdout, ret);
      if(tick == 5000)
      {
        if(debug)
          printf("Waiting for external trigger, point = %d, tick=%d\n", point, tick);
	point++;
	tick = 0;
      } 
      if(point == 10)
      {
        printf("point = %d, tick=%d\n", point,tick);

        ret = U2F_Reg_Write(O_RDOMOD, 0x0); 
        if (ret)
          rcc_error_print(stdout, ret);

        ret = U2F_Reg_Read(O_RDOMOD, &rdata); 
        if (ret)
          rcc_error_print(stdout, ret);
        if(debug)
          printf("External Trigger Disabled = %x\n", rdata);

        return(0);  		
      }
    }
  }


  /*************************************************************************/
  /*			Perform the Readout				   */
  /*************************************************************************/
 
  /***  Read EVWORD Register  ***/
  ret = U2F_Reg_Read(O_EVWORD, &rdata); 
  if (ret)
    rcc_error_print(stdout, ret);
  if(debug)
    printf("EVWORD = %x\n", rdata);

  if((rdata & 0x00000010) && (rdata & 0x00000008)) 
  {
    if(debug)
    {
      printf("STARTING EVENT READOUT\n");  	
      if(rdata & 0x00000004)
  	printf("EVENT READOUT WAS CONCLUDED WITHOUT DATA TRANSFER\n");
    }

    /*** Calculating Data Block Size ***/  
    float_nwords =  (nsamples + 2) / 4.0;
    frac_nwords = modf(float_nwords, &int_nwords);
    if(frac_nwords)
      rsize = int_nwords + 2;
    else
      rsize = int_nwords +1; 
    if(debug)
      printf("\nData Block Size = %d\n", rsize);

    while(!(rdata & 0x00000004))
    {
      /*******************************/
      /***  Read EVWORD Register  ***/
      /******************************/
      ret = U2F_Reg_Read(O_EVWORD, &rdata); 
      if (ret)
        rcc_error_print(stdout, ret);

      /*******************************/
      /***  Reading Data  	   ***/
      /*******************************/

      /*** Reading DMEM1  ***/

      if (rdata & 0x00000002)
      {
	if(debug)
	  printf("\n\nReading Channel #%d from DMEM1\n", chnum);

	ret = U2F_DM1_Read(rsize, &osize, 0, memdata);
	if (ret)
	  rcc_error_print(stdout, ret);

	for (loop = 0; loop < osize; loop++) 
	{
	  fprintf(file, "%d\n", memdata[loop * 4]);
	  fprintf(file, "%d\n", memdata[loop * 4 +1]);
	  fprintf(file, "%d\n", memdata[loop * 4 +2]);
	  fprintf(file, "%d\n", memdata[loop * 4 +3]);	
	}

	/* Read EVWORD */ 
	ret = U2F_Reg_Read(O_EVWORD, &rdata); 
	if (ret)
	  rcc_error_print(stdout, ret);
	if(debug)
	  printf("EVWORD = %x\n", rdata);

	/***  Reset DMEM1 flag    ***/
	ret = U2F_Exec_Command(C_RS_DMEM1, 0);
	if (ret)
	  rcc_error_print(stdout, ret);

	/* Read EVWORD */ 
	ret = U2F_Reg_Read(O_EVWORD, &rdata); 
	if (ret)
	  rcc_error_print(stdout, ret);
	if(debug)
	  printf("EVWORD = %x\n", rdata);
      }

      /***  Reading DMEM2  ***/
      if (rdata & 0x00000001)
      {
	chnum++;
	if(debug)	
	  printf("\n\nReading Channel #%d from DMEM2\n", chnum);

	ret = U2F_DM2_Read(rsize, &osize, 0, memdata); 
	if (ret)
	  rcc_error_print(stdout, ret);

	for (loop = 0; loop < osize; loop++) 
	{
	  fprintf(file, "%d\n", memdata[loop * 4]);
	  fprintf(file, "%d\n", memdata[loop * 4 +1]);
	  fprintf(file, "%d\n", memdata[loop * 4 +2]);
	  fprintf(file, "%d\n", memdata[loop * 4 +3]);	
	}

	/* Read EVWORD */ 
	ret = U2F_Reg_Read(O_EVWORD, &rdata); 
	if (ret)
	  rcc_error_print(stdout, ret);
	if(debug)
	  printf("EVWORD = %x\n", rdata);

	/***  Reset DMEM2 flag    ***/
	ret = U2F_Exec_Command(C_RS_DMEM2, 0);
	if (ret)
	  rcc_error_print(stdout, ret);

	/* Read EVWORD */ 
	ret = U2F_Reg_Read(O_EVWORD, &rdata); 
	if (ret)
	  rcc_error_print(stdout, ret);
	if(debug)
	  printf("EVWORD = %x\n", rdata);
      }
      chnum++;
    }

    /*** Reset EVWORD ***/
    ret = U2F_Exec_Command(C_TRG_CLR, 0);
    if (ret)
      rcc_error_print(stdout, ret);

    /*** Read EVWORD ***/ 
    ret = U2F_Reg_Read(O_EVWORD, &rdata); 
    if (ret)
      rcc_error_print(stdout, ret);
    if(debug)
      printf("EVWORD = %x\n", rdata);
  }
  printf("Event Nr. %d completed\n", nev);
}
  fclose(file); 

  ret = U2F_Reg_Write(O_RDOMOD, 0x0); 
  if (ret)
    rcc_error_print(stdout, ret);

  ret = U2F_Reg_Read(O_RDOMOD, &rdata); 
  if (ret)
    rcc_error_print(stdout, ret);
  if(debug)
    printf("External Trigger Disabled = %x\n", rdata);
    
  return ret;
}
