/************************************************************************/
/*									*/
/*  This is the U2F library 	   				        */
/*									*/
/*  24. Jun. 04  MAJO  created						*/
/*									*/
/*******C 2004 - The software with that certain something****************/

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <fcntl.h>
#include "u2f.h"


/***********/
/* Globals */
/***********/
static u_int init = 0, u2f_is_open = 0;
extern u_int traceLevel, packageId;
int emode;
u2f_bank_t u2f_imem, u2f_rmem, u2f_pmem, u2f_acl, u2f_dmh1, u2f_dml1, u2f_dmh2, u2f_dml2, u2f_registers;

/***********************************************/
u_int U2F_banks_alloc(int mode, u2f_bank_t* bank) 
/***********************************************/
{
  if (mode == HW)
    return(0);
  else
  {  
    if (!bank) 
      return(RCC_ERROR_RETURN(0, U2F_SUCCESS));

    bank->data = (u_int*)malloc(sizeof(u_int)*bank->size);
    if (!bank->data) 
    {
      DEBUG_TEXT(P_ID_U2F, 5 ,("U2F_banks_alloc: Failed to allocate memory for data bank\n"));
      return(RCC_ERROR_RETURN(0, U2F_MALLOC));
    }
  }
  return(0);
}


/****************************/
u_int U2F_banks_init(int mode)
/****************************/
{
  int ret;
  
  if (init > 0) 
  {
    init++;
    return(RCC_ERROR_RETURN(0, U2F_SUCCESS));
  }
  
  ret = U2F_banks_alloc(mode, &u2f_imem);
  if (ret)
    return(RCC_ERROR_RETURN(0, U2F_MALLOC));

  ret = U2F_banks_alloc(mode, &u2f_pmem);
  if (ret)
    return(RCC_ERROR_RETURN(0, U2F_MALLOC));

  ret = U2F_banks_alloc(mode, &u2f_rmem);
  if (ret)
    return(RCC_ERROR_RETURN(0, U2F_MALLOC));

  ret = U2F_banks_alloc(mode, &u2f_acl);
  if (ret)
    return(RCC_ERROR_RETURN(0, U2F_MALLOC));

  ret = U2F_banks_alloc(mode, &u2f_dmh1);
  if (ret)
    return(RCC_ERROR_RETURN(0, U2F_MALLOC));

  ret = U2F_banks_alloc(mode, &u2f_dml1);
  if (ret)
    return(RCC_ERROR_RETURN(0, U2F_MALLOC));

  ret = U2F_banks_alloc(mode, &u2f_dmh2);
  if (ret)
    return(RCC_ERROR_RETURN(0, U2F_MALLOC));

  ret = U2F_banks_alloc(mode, &u2f_dml2);
  if (ret)
    return(RCC_ERROR_RETURN(0, U2F_MALLOC));

  ret = U2F_banks_alloc(mode, &u2f_registers);
  if (ret)
    return(RCC_ERROR_RETURN(0, U2F_MALLOC));

  init = 1;
  return(RCC_ERROR_RETURN(0, U2F_SUCCESS));
}


/******************************/
void U2F_banks_destroy(int mode)
/******************************/
{
  if (--init != 0)
    return;

  if (mode == HW)
    return;
  else
  {
    free(u2f_imem.data);
    free(u2f_pmem.data);
    free(u2f_rmem.data);
    free(u2f_acl.data);
    free(u2f_dmh1.data);
    free(u2f_dml1.data);
    free(u2f_dmh2.data);
    free(u2f_dml2.data);
    free(u2f_registers.data);
  }
}


/**********************************/
u_int U2F_Open(int mode, char *node)
/**********************************/
{  
  u_int ret;
 
  if (u2f_is_open) 
  {
    u2f_is_open++;             /* keep track of multiple open calls */
    return(RCC_ERROR_RETURN(0, U2F_SUCCESS));
  }

  emode = mode;
  u2f_imem.id    = A_IMEM;
  u2f_imem.size  = S_IMEM;
  u2f_imem.mask1 = 0xffff;
  u2f_imem.mask2 = 0x00ff;
  u2f_imem.data  = 0;

  u2f_pmem.id    = A_PMEM;
  u2f_pmem.size  = S_PMEM;
  u2f_pmem.mask1 = 0xffff;
  u2f_pmem.mask2 = 0x0;
  u2f_pmem.data  = 0;

  u2f_rmem.id    = A_RMEM;
  u2f_rmem.size  = S_RMEM;
  u2f_rmem.mask1 = 0xffff;
  u2f_rmem.mask2 = 0x000f;
  u2f_rmem.data  = 0;

  u2f_acl.id     = A_ACL;
  u2f_acl.size   = S_ACL;
  u2f_acl.mask1  = 0xffff;
  u2f_acl.mask2  = 0x0;
  u2f_acl.data   = 0;

  u2f_dmh1.id    = A_DMH1;
  u2f_dmh1.size  = S_DM;
  u2f_dmh1.mask1 = 0xffff;
  u2f_dmh1.mask2 = 0x000f;
  u2f_dmh1.data  = 0;

  u2f_dml1.id    = A_DML1;
  u2f_dml1.size  = S_DM;
  u2f_dml1.mask1 = 0xffff;
  u2f_dml1.mask2 = 0x000f;
  u2f_dml1.data  = 0;

  u2f_dmh2.id    = A_DMH2;
  u2f_dmh2.size  = S_DM;
  u2f_dmh2.mask1 = 0xffff;
  u2f_dmh2.mask2 = 0x000f;
  u2f_dmh2.data  = 0;

  u2f_dml2.id    = A_DML2;
  u2f_dml2.size  = S_DM;
  u2f_dml2.mask1 = 0xffff;
  u2f_dml2.mask2 = 0x000f;
  u2f_dml2.data  = 0;

  u2f_registers.id    = A_REGISTERS;
  u2f_registers.size  = O_RDOMOD + 1;
  u2f_registers.mask1 = 0xffff;
  u2f_registers.mask2 = 0xffff;
  u2f_registers.data  = 0;

  /* open the error package */
  ret = rcc_error_init((err_pid)P_ID_U2F, U2F_err_get);
  if (ret) 
  {
    DEBUG_TEXT(P_ID_U2F, 5 ,("U2F_Open: Failed to open error package\n"));
    return(RCC_ERROR_RETURN(0, U2F_ERROR_FAIL)); 
  }
  DEBUG_TEXT(P_ID_U2F, 20 ,("U2F_Open: error package opened\n"));

  /* setup the instruction memory bank information */
  U2F_banks_init(mode);

  if (mode == HW)
  {
    ret = ALTRO_Open(node);
    if (ret) 
    {
      DEBUG_TEXT(P_ID_U2F, 5 ,("U2F_Open: Failed to open the ALTRO library\n"));
      return(RCC_ERROR_RETURN(ret, U2F_ALTRO_FAIL)); 
    }
    DEBUG_TEXT(P_ID_U2F, 10 ,("U2F_Open: ALTRO library opened\n"));
  }
  
  u2f_is_open = 1;
  
  return(U2F_SUCCESS);
}


/*******************/
u_int U2F_Close(void)
/*******************/
{
  u_int ret = 0;
  
  U2FISOPEN;
    
  if (u2f_is_open > 1) 
    u2f_is_open--;
  else 
  {  
    if (emode == HW)
    {
      ret = ALTRO_Close();
      if (ret) 
      {
	DEBUG_TEXT(P_ID_U2F, 5,("U2F_Close: Failed to close the ALTRO library\n"));
	return(RCC_ERROR_RETURN(ret, U2F_ALTRO_FAIL)); 
      }
      DEBUG_TEXT(P_ID_U2F, 10 ,("U2F_Close: ALTRO library closed\n"));
    }
    U2F_banks_destroy(emode);
    u2f_is_open = 0;
  }
  
  return(U2F_SUCCESS);
}


/**************************************************************************************************/
static u_int U2F_Memory_Read(u2f_bank_t* bank, u_int isize, u_int* osize, u_int offset, u_int* data)
/**************************************************************************************************/
{
  altro_bulk_out_t out;
  altro_bulk_in_t in;
  u_short *msg;
  u_int ret, msize, loop, *ptr;
  err_str pid, code;

  U2FISOPEN;

  DEBUG_TEXT(P_ID_U2F, 20 ,("U2F_Memory_Read: bank = 0x%04x isize = %d\n", bank->id, isize));

  msize = isize * 2;    /* one word in the IMEM takes two 16 bit words in the message */
 
  if ((isize + offset) > bank->size) 
  {    
    DEBUG_TEXT(P_ID_U2F, 5 , ("U2F_Memory_Read: Parameter isize is out of range [0,%d]\n", bank->size));
    return(RCC_ERROR_RETURN(0, U2F_RANGE));
  } 

  if (emode == SW)
  {
    DEBUG_TEXT(P_ID_U2F, 20 , ("U2F_Memory_Read: Emulation: isize      = %d\n", isize));
    DEBUG_TEXT(P_ID_U2F, 20 , ("U2F_Memory_REad: Emulation: offset     = %d\n", offset));
    DEBUG_TEXT(P_ID_U2F, 20 , ("U2F_Memory_Read: Emulation: bank->id   = %d\n", bank->id));
    DEBUG_TEXT(P_ID_U2F, 20 , ("U2F_Memory_Read: Emulation: bank->size = %d\n", bank->size));
    *osize = isize;
    for (loop = 0; loop < isize; loop++)
      data[loop] = bank->data[loop + offset];     
    return(U2F_SUCCESS);
  }

  /* Construct a message - First 4 bits are the instruction, and one word in a block takes two 16 bit words in the message */
  msg        = (u_short *)&out.data[0];
  msg[0]     = (isize << 5) + I_READ; 
  msg[1]     = bank->id + offset;
  out.nbytes = 4;

  DEBUG_TEXT(P_ID_U2F, 20 ,("U2F_Memory_Read: msg[0] = 0x%08x\n", msg[0]));
  DEBUG_TEXT(P_ID_U2F, 20 ,("U2F_Memory_Read: msg[1] = 0x%08x\n", msg[1]));

  /* Tell the USB2FEC that we want to read the a block */
  ret = ALTRO_Send(&out);
  if (ret) 
  {    
    err_str pid, code;
    ALTRO_err_get(ret, pid, code);
    DEBUG_TEXT(P_ID_U2F, 5 , ("U2F_Memory_Read: Error from ALTRO_Send %s %s\n", pid, code));
    return(RCC_ERROR_RETURN(ret, U2F_ALTRO_FAIL)); 
  }    

  /* Now read the data */
  ret = ALTRO_Get(&in);
  if (ret) 
  {
    ALTRO_err_get(ret, pid, code);
    DEBUG_TEXT(P_ID_U2F, 5 ,("U2F_Memory_Read: Error (%d) from ALTRO_Get: %s\n", ret, code));
    return(RCC_ERROR_RETURN(ret, U2F_ALTRO_FAIL)); 
  }

  DEBUG_TEXT(P_ID_U2F, 20 , ("U2F_Memory_Read: %d bytes received from the USB2FEC\n", in.nbytes));
  if (in.nbytes != (msize * 2)) 
  {
    DEBUG_TEXT(P_ID_U2F, 5 , ("U2F_Memory_Read: Wrong number of bytes received: was %d, should be %d\n", in.nbytes, msize * 2));
    *osize = 0;
    return(RCC_ERROR_RETURN(ret, U2F_ODD_BYTES)); 
  }

  /* We got all we needed */
  *osize = isize;
  ptr    = (u_int *) &in.data[0];
  /* copy the data from the altro_bulk_in_t structure to the data array */

  for(loop = 0; loop < *osize; loop++) 
  { 
    data[loop] = ptr[loop];
    DEBUG_TEXT(P_ID_U2F, 30 , ("\tData # %3d:\t0x%08x 0x%04x 0x%04x\n", loop, data[loop], in.data[2*loop], in.data[2*loop+1]));
  }
  
  return(U2F_SUCCESS);
}


/************************************************************************/
u_int U2F_IMEM_Read(u_int isize, u_int *osize, u_int offset, u_int data[])
/************************************************************************/
{
  int ret;
  
  DEBUG_TEXT(P_ID_U2F, 20 ,("U2F_IMEM_Read: isize = %d\n", isize));
  ret = U2F_Memory_Read(&u2f_imem, isize, osize, offset, data);
  return ret;
}


/************************************************************************/
u_int U2F_PMEM_Read(u_int isize, u_int *osize, u_int offset, u_int data[])
/************************************************************************/  
{
  int ret;
  
  DEBUG_TEXT(P_ID_U2F, 20 ,("U2F_PMEM_Read: isize = %d\n", isize));
  if (isize > B_PMEM)  
  {    
    printf("\nNumber of words cannot exceed 256 words\n"); 
    DEBUG_TEXT(P_ID_U2F, 5 , ("U2F_PMEM_Read: Parameter isize is out of range\n"));
    return(RCC_ERROR_RETURN(0, U2F_RANGE));
  }   
  ret = U2F_Memory_Read(&u2f_pmem, isize, osize, offset, data);
  return ret;
}
		
  
/************************************************************************/
u_int U2F_RMEM_Read(u_int isize, u_int *osize, u_int offset, u_int data[])
/************************************************************************/
{
  int ret;
  
  DEBUG_TEXT(P_ID_U2F, 20 ,("U2F_IMEM_Read: isize = %d\n", isize));
  ret = U2F_Memory_Read(&u2f_rmem, isize, osize, offset, data);
  return ret;
}

  
/***********************************************************************/
u_int U2F_ACL_Read(u_int isize, u_int *osize, u_int offset, u_int data[])
/***********************************************************************/
{
  int ret;
  DEBUG_TEXT(P_ID_U2F, 20 ,("U2F_IMEM_Read: isize = %d\n", isize));
  ret = U2F_Memory_Read(&u2f_acl, isize, osize, offset, data);
  return ret;
}


/***********************************************************************************/
u_int U2F_DM_Read(u_int which, u_int isize, u_int *osize, u_int offset, u_int data[])
/***********************************************************************************/
{
  static u_int low[2048], high[2048];
  u_int losize, hosize, ret, loop;
  
  u2f_bank_t* lbank = 0;
  u2f_bank_t* hbank = 0;
  switch (which) 
  {
    case 1:
      lbank = &u2f_dml1;
      hbank = &u2f_dmh1;
      break;
    case 2:
      lbank = &u2f_dml2;
      hbank = &u2f_dmh2;
      break;
    default:
      DEBUG_TEXT(P_ID_U2F, 5, ("Unknown bank #:%d\n", which));
      return(RCC_ERROR_RETURN(0, U2F_RANGE));
  }
  
  ret = U2F_Memory_Read(lbank, isize, &losize, offset, low);
  if (ret)
  {
   *osize = 0;
   return ret;
  }
  
  ret = U2F_Memory_Read(hbank, isize, &hosize, offset, high);
  if (ret)
  {
   *osize = 0;
   return ret;
  }
  
  if (losize != hosize) 
  {
    DEBUG_TEXT(P_ID_U2F, 5 , ("U2F_DM_Read: Inconsistent number of bytes received\n"));
    *osize = 0;
    return(RCC_ERROR_RETURN(ret, U2F_INC_BYTES)); 
  }

  if (losize != isize) 
  {
    DEBUG_TEXT(P_ID_U2F, 5 , ("U2F_DM%d_Read: Wrong number (%d) of words received for the low bits, should be %d\n", which, losize, isize));
    *osize = 0;
    return(RCC_ERROR_RETURN(ret, U2F_ODD_BYTES)); 
  }
  
  if (hosize != isize) 
  {
    DEBUG_TEXT(P_ID_U2F, 5 , ("U2F_DM%d_Read: Wrong number (%d) of words received for the hight bits, should be %d\n", which, hosize, isize));
    *osize = 0;
    return(RCC_ERROR_RETURN(ret, U2F_ODD_BYTES)); 
  }
  
  *osize = isize;
  /* copy the data from the altro_bulk_in_t structures to the data array */
  for(loop = 0; loop < *osize; loop++)
  {
    data[4*loop]   = low[loop]          & 0x3ff;
    data[4*loop+1] = (low[loop] >> 10)  & 0x3ff;
    data[4*loop+2] = high[loop]         & 0x3ff;
    data[4*loop+3] = (high[loop] >> 10) & 0x3ff;
  }
  
  return(U2F_SUCCESS); 
}


/***********************************************************************/
u_int U2F_DM1_Read(u_int isize, u_int *osize, u_int offset, u_int data[])
/***********************************************************************/
{
  return U2F_DM_Read(1, isize, osize, offset, data);
}


/***********************************************************************/
u_int U2F_DM2_Read(u_int isize, u_int *osize, u_int offset, u_int data[])
/***********************************************************************/
{
  return U2F_DM_Read(2, isize, osize, offset, data);
}


/****************************************/
u_int U2F_Reg_Read(u_int reg, u_int *data)
/****************************************/
{
  u_int osize;

  DEBUG_TEXT(P_ID_U2F, 10, ("U2F_Reg_Read: Function called for reg = %d\n", reg));
  if ((reg != O_ERRST)  && (reg != O_TRCFG)  && (reg != O_TRCNT) && (reg != O_IRADD) 
   && (reg != O_IRDAT)  && (reg != O_EVWORD) && (reg != O_LWADD) && (reg != O_ACTFEC) 
   && (reg != O_FMIREG) && (reg != O_FMOREG) && (reg != O_RDOMOD))
  {    
    DEBUG_TEXT(P_ID_U2F, 5, ("U2F_Reg_Read: Unknown register %d\n", reg));
    return(RCC_ERROR_RETURN(0, U2F_RANGE)); 
  }
  return U2F_Memory_Read(&u2f_registers, 1, &osize, reg, data);
}  


/********************************/
u_int U2F_FMIREG_Read(u_int* data)
/********************************/
{
  return U2F_Reg_Read(O_FMIREG, data);
}


/********************************/
u_int U2F_FMOREG_Read(u_int* data)
/********************************/
{
  return U2F_Reg_Read(O_FMOREG, data);
}


/*******************************/
u_int U2F_ERRST_Read(u_int* data)
/*******************************/
{
  return U2F_Reg_Read(O_ERRST, data);
}


/*******************************/
u_int U2F_TRCFG_Read(u_int* data)
/*******************************/
{
  return U2F_Reg_Read(O_TRCFG, data);
}


/*******************************/
u_int U2F_TRCNT_Read(u_int* data)
/*******************************/
{
  return U2F_Reg_Read(O_TRCNT, data);
}


/*******************************/
u_int U2F_LWADD_Read(u_int* data)
/*******************************/
{
  return U2F_Reg_Read(O_LWADD, data);
}


/*******************************/
u_int U2F_IRADD_Read(u_int* data)
/*******************************/
{
  return U2F_Reg_Read(O_IRADD, data);
}


/*******************************/
u_int U2F_IRDAT_Read(u_int* data)
/*******************************/
{
  return U2F_Reg_Read(O_IRDAT, data);
}


/********************************/
u_int U2F_EVWORD_Read(u_int* data)
/********************************/
{
  return U2F_Reg_Read(O_EVWORD, data);
}

/********************************/
u_int U2F_ACTFEC_Read(u_int* data)
/********************************/
{
  return U2F_Reg_Read(O_ACTFEC, data);
}

/********************************/
u_int U2F_RDOMOD_Read(u_int* data)
/********************************/
{
  return U2F_Reg_Read(O_RDOMOD, data);
}


/*****************************************************/
u_int U2F_ERRST_decode(u_int data, errst_t *parameters)
/*****************************************************/
{
  U2FISOPEN;

  parameters->pattern_error = data & 0x1; 
  parameters->abort         = (data >> 1) & 0x1;
  parameters->timeout       = (data >> 2) & 0x1;
  parameters->altro_error   = (data >> 3) & 0x1;
  
  return(U2F_SUCCESS);
}


/*****************************************************/
u_int U2F_TRCFG_decode(u_int data, trcfg_t *parameters)
/*****************************************************/
{
  U2FISOPEN;

  parameters->tw    =data & 0x3fff;
  parameters->bmd   =(data >> 14) & 0x1;
  parameters->mode  =(data >> 15) & 0x3;
  parameters->pop   =(data >> 18) & 0x1;
  parameters->remb  =(data >> 17) & 0xf;
  parameters->empty =(data >> 21) & 0x1;
  parameters->full  =(data >> 22) & 0x1;
  parameters->rd_pt =(data >> 23) & 0x7;
  parameters->wr_pt =(data >> 26) & 0x7;

  return(U2F_SUCCESS);
}
  
  
/*****************************************************/
u_int U2F_TRCNT_decode(u_int data, trcnt_t *parameters)
/*****************************************************/
{
  U2FISOPEN;

  parameters->nta = data & 0xffff;
  parameters->ntr = (data >> 16) & 0xffff;

  return(U2F_SUCCESS);
}


/**************************************/
u_int U2F_ERRST_Get(errst_t *parameters) 
/**************************************/
{
  u_int data, ret;
  
  ret = U2F_ERRST_Read(&data);
  if (ret) 
    return ret;
  return U2F_ERRST_decode(data, parameters);
}


/**************************************/
u_int U2F_TRCFG_Get(trcfg_t *parameters) 
/**************************************/
{
  u_int data, ret;
  
  ret = U2F_TRCFG_Read(&data);
  if (ret) return ret;
  return U2F_TRCFG_decode(data, parameters);
}


/**************************************/
u_int U2F_TRCNT_Get(trcnt_t *parameters) 
/**************************************/
{
  u_int data, ret;

  ret = U2F_TRCNT_Read(&data);
  if (ret) return ret;
  return U2F_TRCNT_decode(data, parameters);
}


/**************************************************************************************/
static u_int U2F_Memory_Write(u2f_bank_t* bank, u_int osize, u_int offset, u_int data[])
/**************************************************************************************/
{
  altro_bulk_out_t out;
  u_short *msg, ptr;
  u_int ret, msize, loop;
  
  U2FISOPEN;
  DEBUG_TEXT(P_ID_U2F, 10 ,("U2F_Memory_Write: bank = 0x%04x  offset = 0x%04x  osize = %d\n", bank->id, offset, osize));

  /* one word in the IMEM takes two 16 bit words in the message */
  msize = osize * 2;  
  if ((osize + offset) > bank->size) 
  {    
    DEBUG_TEXT(P_ID_U2F, 5 , ("U2F_Memory_Write: Parameter osize (%d [%d]) is out of range [0,%d)\n", osize, offset, bank->size));
    return(RCC_ERROR_RETURN(0, U2F_RANGE));
  } 
 
  if (emode == SW)
  {
    DEBUG_TEXT(P_ID_U2F, 20 , ("U2F_Memory_Write: Emulation: osize      = %d\n", osize));
    DEBUG_TEXT(P_ID_U2F, 20 , ("U2F_Memory_Write: Emulation: offset     = %d\n", offset));
    DEBUG_TEXT(P_ID_U2F, 20 , ("U2F_Memory_Write: Emulation: bank->id   = 0x%0x4\n", bank->id));
    DEBUG_TEXT(P_ID_U2F, 20 , ("U2F_Memory_Write: Emulation: bank->size = %d\n", bank->size));
    for (loop = 0; loop < osize; loop++)
       bank->data[loop + offset] = data[loop];     
    return(U2F_SUCCESS);
  }
 
  /* Construct a message */
  msg    = (u_short *)&out.data[0];
  msg[0] = (osize << 5) + I_WRITE; 
  msg[1] = bank->id + offset;
  ptr    = 2;
  for(loop = 0; loop < osize; loop++) 
  {
    msg[ptr]   = data[loop] & bank->mask1;
    msg[ptr+1] = (bank->mask2 ? (data[loop] >> 16) & bank->mask2 : 0);
    DEBUG_TEXT(P_ID_U2F, 30 , ("\tData # %3d:\tx%08x 0x%04x 0x%04x\n", loop, data[loop], msg[ptr], msg[ptr+1]));
    ptr += 2;
  } 
  out.nbytes = (2 + msize) * 2;  /*  16 bit word to byte */

  /* Tell the USB2FEC that we want to write the memory bank */ 
  ret = ALTRO_Send(&out);
  if (ret) 
  {      
    DEBUG_TEXT(P_ID_U2F, 5 ,("U2F_Memory_Write: Error from ALTRO_Send\n"));
    return(RCC_ERROR_RETURN(ret, U2F_ALTRO_FAIL)); 
  }    
  DEBUG_TEXT(P_ID_U2F, 20 ,("U2F_Memory_Write: ret = %d\n", ret));

  return(U2F_SUCCESS);  
}


/***********************************************************/
u_int U2F_IMEM_Write(u_int osize, u_int offset, u_int data[])
/***********************************************************/
{
  u_int ret;
  
  DEBUG_TEXT(P_ID_U2F, 10 ,("U2F_IMEM_Write: osize=%d\n", osize));
  ret = U2F_Memory_Write(&u2f_imem, osize, offset, data);
  return ret;
}


/************************************************************/
u_int U2F_PMEM_Write (u_int osize, u_int offset, u_int data[])
/************************************************************/
{
  u_int ret;

  DEBUG_TEXT(P_ID_U2F, 10 ,("U2F_PMEM_Write: osize=%d\n", osize));
  if (osize  > B_PMEM) 
  {    
    printf("\nNumber of words cannot exceed 256 words\n"); 
    DEBUG_TEXT(P_ID_U2F, 5 , ("U2F_PMEM_Write: Parameter isize is out of range\n"));
    return(RCC_ERROR_RETURN(0, U2F_RANGE));
  }  
  ret = U2F_Memory_Write(&u2f_pmem, osize, offset, data);
  return ret;
}

  
/************************************************************/
u_int U2F_RMEM_Write (u_int osize, u_int offset, u_int data[])
/************************************************************/
{
  u_int ret;
  
  DEBUG_TEXT(P_ID_U2F, 10 ,("U2F_IMEM_Write: osize=%d\n", osize));
  ret = U2F_Memory_Write(&u2f_rmem, osize, offset, data);
  return ret;
}

/************************************************************/
u_int U2F_ACL_Write (u_int osize, u_int offset, u_int data[])
/************************************************************/
{
  u_int ret;
  
  DEBUG_TEXT(P_ID_U2F, 10 ,("U2F_IMEM_Write: osize=%d\n", osize));
  ret = U2F_Memory_Write(&u2f_acl, osize, offset, data);
  return ret;
}


/***********************************************************************/
u_int U2F_DM_Write (u_int which, u_int osize, u_int offset, u_int data[])
/***********************************************************************/
{
  static u_int low[2048], high[2048];
  u_int ret, loop, ptr1, ptr2, ptr3;
  
  u2f_bank_t* lbank = 0;
  u2f_bank_t* hbank = 0;
  switch (which) 
  {
    case 1:
      lbank = &u2f_dml1;
      hbank = &u2f_dmh1;
      break;
    case 2:
      lbank = &u2f_dml2;
      hbank = &u2f_dmh2;
      break;
    default:
      DEBUG_TEXT(P_ID_U2F, 5, ("Unknown bank #:%d\n", which));
      return(RCC_ERROR_RETURN(0, U2F_RANGE));
  }

  DEBUG_TEXT(P_ID_U2F, 20, ("Writing to DM%d of size %d starting at %d\n", which, osize, offset));
  ptr1 = 2;
  ptr2 = 2;
  ptr3 = 0;
  for (loop = 0; loop < osize; loop++) 
  {
    low[ptr1++]  = data[ptr3]         & 0xffff;
    low[ptr1++]  = (data[ptr3] >> 16) & 0x000f;
    ptr3++;
    high[ptr2++] = data[ptr3]         & 0xffff;
    high[ptr2++] = (data[ptr3] >> 16) & 0x000f;
    ptr3++;
  }  
 
  ret = U2F_Memory_Write(lbank, osize, offset, low);
  ret = U2F_Memory_Write(hbank, osize, offset, high);
  
  return ret;
}

  
/***********************************************************/
u_int U2F_DM1_Write (u_int osize, u_int offset, u_int data[])
/***********************************************************/
{
  return U2F_DM_Write(1, osize, offset, data);
}


/***********************************************************/
u_int U2F_DM2_Write (u_int osize, u_int offset, u_int data[])
/***********************************************************/
{
  return U2F_DM_Write(2, osize, offset, data);
}


/****************************************/
u_int U2F_Reg_Write(u_int reg, u_int data)
/****************************************/
{
  int ret;
  DEBUG_TEXT(P_ID_U2F, 20 , ("U2F_Reg_Write: reg = %d  data = %d\n", reg, data));
  
  if ((reg != O_TRCFG) && (reg != O_ACTFEC) && (reg != O_FMIREG) && (reg != O_FMOREG) && (reg != O_RDOMOD)) 
  {    
    DEBUG_TEXT(P_ID_U2F, 5 , ("U2F_Reg_Write: The parameter reg is out of range (reg=%d)\n", reg));
    return(RCC_ERROR_RETURN(0, U2F_RANGE)); 
  } 
  ret = U2F_Memory_Write(&u2f_registers, 1, reg, &data);
  
  return(U2F_SUCCESS);
}


/*******************************/
u_int U2F_TRCFG_Write(u_int data)
/*******************************/
{
  return U2F_Reg_Write(O_TRCFG, data);
}


/********************************/
u_int U2F_ACTFEC_Write(u_int data)
/********************************/
{
  return U2F_Reg_Write(O_ACTFEC, data);
}


/*****************************************************/
u_int U2F_TRCFG_encode(u_int *data, trcfg_t parameters)
/*****************************************************/
{
  U2FISOPEN;

  if (parameters.tw > 0x3fff) 
  {    
    DEBUG_TEXT(P_ID_U2F, 5 , ("U2F_TRCFG_encode: Parameter tw is out of range\n"));
    return(RCC_ERROR_RETURN(0, U2F_RANGE));
  }       
  
  if (parameters.bmd > 2) 
  {    
    DEBUG_TEXT(P_ID_U2F, 5 , ("U2F_TRCFG_encode: Parameter bmd is out of range\n"));
    return(RCC_ERROR_RETURN(0, U2F_RANGE));
  }     

  if ((parameters.mode > 3) || (parameters.mode == 1)) 
  {    
    DEBUG_TEXT(P_ID_U2F, 5 , ("U2F_TRCFG_encode: Parameter mode is out of range\n"));
    return(RCC_ERROR_RETURN(0, U2F_RANGE));
  }  

  *data = parameters.tw + (parameters.bmd << 14) + (parameters.mode << 15) + 0x40000;
  
  DEBUG_TEXT(P_ID_U2F, 20 , ("U2F_TRCFG_encode: The encoded data word is %d\n", *data));
 
  return(U2F_SUCCESS);
}

  
/*************************************/
u_int U2F_TRCFG_Set(trcfg_t parameters)
/*************************************/
{
  u_int data, ret;
  
  ret = U2F_TRCFG_encode(&data, parameters);
  if (ret) return ret;
  return U2F_TRCFG_Write(data);
}

  
/*******************************************************/
u_int U2F_File_Read(char *name, u_int size, u_int data[])
/*******************************************************/
{
  u_int loop;
  FILE *inf;
  
  U2FISOPEN;

  /* open the input file */
  inf = fopen(name, "r");
  if (inf == 0) 
  {
    DEBUG_TEXT(P_ID_U2F, 20 ,("U2F_File_Read:  Can't open input file\n"));
    return(RCC_ERROR_RETURN(0, U2F_FILE));
  }

  /* read  the file */
  for (loop = 0; loop < size; loop++) 
  {
    fscanf(inf,"%x",&data[loop]);
    if (feof(inf)) 
    {  
      DEBUG_TEXT(P_ID_U2F, 20 ,("U2F_File_Read: The file is too short\n"));
      return(RCC_ERROR_RETURN(0, U2F_NODATA));
    }
  }
  fclose(inf);
  return(U2F_SUCCESS);
}

  
/********************************************************/
u_int U2F_File_Write(char *name, u_int size, u_int data[])
/********************************************************/
{
  u_int loop;
  FILE *outf;

  U2FISOPEN;

  outf = fopen(name, "a+");
  if (outf==0) 
  {
    DEBUG_TEXT(P_ID_U2F, 5 ,("U2F_File_Write:  Can't open output file\n"));
    return(RCC_ERROR_RETURN(0, U2F_FILE));
  }
  
  /* write  the file */
  for (loop = 0; loop < size; loop++)
    fprintf(outf,"%x\n",data[loop]);

  fclose(outf);
  return(U2F_SUCCESS);  
}


/***********************************************/
u_int U2F_Exec_Command(u_int command, u_int data)
/***********************************************/
{
  static u_int dmem1cnt = 0;
  static u_int dmem2cnt = 0;
  altro_bulk_out_t out;
  u_short *msg;
  u_int address, i, ret;
    
  U2FISOPEN;
  DEBUG_TEXT(P_ID_U2F, 10 , ("U2F_Exec_Command: Called\n"));

  if ((command != C_RS_STATUS) && (command != C_RS_TRCFG) && (command != C_RS_TRCNT) && 
      (command != C_EXEC)      && (command != C_ABORT)    && (command != C_FECRST) && 
      (command != C_SWTRG)     && (command != C_TRG_CLR)  && (command != C_RS_DMEM1) && 
      (command != C_RS_DMEM2)  && (command != C_WRFM)     && (command != C_RDFM) &&
      (command != C_RDABORT) && (command != C_CLR_EVTAG))
  {    
    DEBUG_TEXT(P_ID_U2F, 5 , ("U2F_Exec_Command:  The parameter <command> is out of range\n"));
    return(RCC_ERROR_RETURN(0, U2F_RANGE));
  } 
  
  if ((command == C_EXEC) && (data > 0xff)) 
  {    
    DEBUG_TEXT(P_ID_U2F, 5 , ("U2F_Exec_Command:  The parameter <data> is out of range\n"));
    return(RCC_ERROR_RETURN(0, U2F_RANGE));
  } 

  address = command;
  if (command == C_EXEC)
    address += data;

  DEBUG_TEXT(P_ID_U2F, 20 , ("U2F_Exec_Command: address = 0x%08x\n", address));

  if (emode == SW)
  {
    DEBUG_TEXT(P_ID_U2F, 20 , ("U2F_Exec_Command: Emulation: command = %d\n", command));
    DEBUG_TEXT(P_ID_U2F, 20 , ("U2F_Exec_Command: Emulation: data    = %d\n", data));
    DEBUG_TEXT(P_ID_U2F, 20 , ("U2F_Exec_Command: command = %d\n", command));
    switch(command)
    {
      case C_RS_STATUS:
        printf("Status reset");
        u2f_registers.data[O_ERRST] = 0x0;
        break;

      case C_RS_TRCFG:
        printf("Trigger config reset");
        u2f_registers.data[O_TRCFG] = 0x0;
        break;
      
      case C_RS_TRCNT:
        printf("Trigger counter reset");
        u2f_registers.data[O_TRCNT] = 0x0;
        break;
      
      case C_EXEC:
	printf("Execute instruction memory at offset %d\n", data);
	for (i = data; i < u2f_imem.size; i++) 
	{
	  printf("%3d  0x%08x\n", i, u2f_imem.data[i]);
	  if (u2f_imem.data[i] == 0x390000) break;
	  if (u2f_imem.data[i] & (1 << 22)) 
	  {
	    if (u2f_imem.data[i] & (1 << 20)) 
	      u2f_registers.data[O_IRDAT] = u2f_imem.data[i];
	    else 
	      u2f_registers.data[O_IRADD] = u2f_imem.data[i];
	  }
	}
	return(ALTRO_SUCCESS);
      
      case C_ABORT:
        printf("Abort current instructions\n");
        break;
      
      case C_FECRST:
        printf("Reset front end cards\n");
        break;
      
      case C_SWTRG:
	printf("Software trigger\n");
	u2f_registers.data[O_EVWORD] =  0x1B;
	u2f_registers.data[O_TRCNT]  += 0x1001;
	u2f_registers.data[O_LWADD]  =  0x4242;
	write_dummy_data(1, 0, 0, 256);
	write_dummy_data(2, 0, 1, 256);
        break;
      
      case C_RS_DMEM1:
	printf("DMEM 1 reset\n");
	dmem1cnt++;
	if (dmem1cnt >= 16) 
	  dmem1cnt = 0;
        break;
      
      case C_RS_DMEM2:
	printf("DMEM 2 reset\n");
	dmem2cnt++;
	if (dmem2cnt >= 16) 
	{
	  dmem2cnt = 0;
	  u2f_registers.data[O_EVWORD] = 0x4;
	}
        break;
      
      case C_TRG_CLR:
        printf("Event word reset");
        u2f_registers.data[O_EVWORD] = 0x0;
        break;
      
      case C_WRFM:
        printf("Write FM");
        break;
      
      case C_RDFM:
        printf("Read FM");
        break;

      case C_RDABORT:
        printf("Abort large block read out");
        break;
    }
    return(U2F_SUCCESS);
  }

  /* Construct a message */
  msg = (u_short *)&out.data[0];
  msg[0] = I_WRITE; 
  msg[1] = address;
  out.nbytes = 4;
   
  /* Send the message  */
  DEBUG_TEXT(P_ID_U2F, 20 , ("U2F_Exec_Command: Calling ALTRO_Send\n"));
  ret = ALTRO_Send(&out);
  if (ret) 
  {      
    DEBUG_TEXT(P_ID_U2F, 5 , ("U2F_Exec_Command: Failed to close the ALTRO library\n"));
    return(RCC_ERROR_RETURN(ret, U2F_ALTRO_FAIL)); 
  }      

  return(U2F_SUCCESS);  
}


/****************************/
u_int U2F_Exec_RS_STATUS(void)
/****************************/
{
  return U2F_Exec_Command(C_RS_STATUS, 0);
} 


/***************************/
u_int U2F_Exec_RS_TRCFG(void)
/***************************/
{
  return U2F_Exec_Command(C_RS_TRCFG, 0);
} 


/***************************/
u_int U2F_Exec_RS_TRCNT(void)
/***************************/
{
  return U2F_Exec_Command(C_RS_TRCNT, 0);
}  


/*****************************/
u_int U2F_Exec_EXEC(u_int data)
/*****************************/
{
  return U2F_Exec_Command(C_EXEC, data);
} 


/********************/
u_int U2F_Exec_ABORT(void)
/********************/
{
  return U2F_Exec_Command(C_ABORT, 0);
} 


/*************************/
u_int U2F_Exec_FECRST(void)
/*************************/
{
  return U2F_Exec_Command(C_FECRST, 0);
}

/************************/
u_int U2F_Exec_SWTRG(void)
/************************/
{
  return U2F_Exec_Command(C_SWTRG, 0);
} 


/**************************/
u_int U2F_Exec_TRG_CLR(void)
/**************************/
{
  return U2F_Exec_Command(C_TRG_CLR, 0);
} 


/***************************/
u_int U2F_Exec_RS_DMEM1(void)
/***************************/
{
  return U2F_Exec_Command(C_RS_DMEM1, 0);
}  


/***************************/
u_int U2F_Exec_RS_DMEM2(void)
/***************************/
{
  return U2F_Exec_Command(C_RS_DMEM2, 0);
}
     

/*******************/
u_int U2F_Reset(void)
/*******************/
{
  altro_bulk_out_t out;
  u_short *msg;
  u_int ret;
  
  U2FISOPEN;
 
  DEBUG_TEXT(P_ID_U2F, 10 ,("U2F_Reset: function called\n"));

  if (emode == SW)
  {
    DEBUG_TEXT(P_ID_U2F, 20 , ("U2F_Reset: Emulation: Nothiong to be done\n"));
    return(0);
  }
    
  /* Construct a message */
  msg = (u_short *)&out.data[0];
  msg[0] = I_GRESET; 
  out.nbytes = 2;

  /* Tell the USB2FEC that we want to reset the card */
  ret = ALTRO_Send(&out);
  if (ret) 
  {      
    DEBUG_TEXT(P_ID_U2F, 5 ,("U2F_Reset: Error from ALTRO_Send\n"));
    return(RCC_ERROR_RETURN(ret, U2F_ALTRO_FAIL)); 
  } 
  return(0);
}


/*********************************************************************/
u_int U2F_ReadOut(u_int bsize, u_int *osize, u_char data[], u_int mode)
/*********************************************************************/
{
  u_int loop, ret = 0, cont, index;
  altro_bulk_in_t in;
  err_str pid, code;
  
  DEBUG_TEXT(P_ID_U2F, 10 ,("U2F_ReadOut: Function called\n"));
  U2FISOPEN;

  if (mode & M_FIRST)
  {
    DEBUG_TEXT(P_ID_U2F, 20 ,("U2F_ReadOut: Enabling read-out mode\n"));
    ret = U2F_Reg_Write(O_RDOMOD, 3);
    if (ret)
    {
      ALTRO_err_get(ret, pid, code);
      DEBUG_TEXT(P_ID_U2F, 5 ,("U2F_ReadOut: Error (%d) from U2F_Reg_Write: %s\n", ret, code));
      return(RCC_ERROR_RETURN(0, ret)); 
    }
  } 

  if (mode & M_TRIGGER)
  {
    DEBUG_TEXT(P_ID_U2F, 10 ,("U2F_ReadOut: Sending S/W trigger\n"));
    ret = U2F_Exec_Command(C_SWTRG, 0);
    if (ret)
    {
      ALTRO_err_get(ret, pid, code);
      DEBUG_TEXT(P_ID_U2F, 5 ,("U2F_ReadOut: Error (%d) from U2F_Exec_Command: %s\n", ret, code));
      return(RCC_ERROR_RETURN(0, ret)); 
    }
  }

  DEBUG_TEXT(P_ID_U2F, 10 ,("U2F_ReadOut: Reading data\n"));
  cont = 1;
  index = 0;
  while(cont)
  {
    if ((index + 1024) > bsize)
    {
      DEBUG_TEXT(P_ID_U2F, 5 ,("U2F_ReadOut: The next USB burst would overflow the data array\n"));
      
      if (mode & M_LAST)
      {
	DEBUG_TEXT(P_ID_U2F, 20 ,("U2F_ReadOut: Aborting read-out mode\n"));
	ret = U2F_Exec_Command(C_RDABORT, 0);
	if (ret)
	{
	  ALTRO_err_get(ret, pid, code);
	  DEBUG_TEXT(P_ID_U2F, 5 ,("U2F_ReadOut: Error (%d) from U2F_Exec_Command: %s\n", ret, code));
	  return(RCC_ERROR_RETURN(0, ret)); 
	}

	DEBUG_TEXT(P_ID_U2F, 20 ,("U2F_ReadOut: Disabling read-out mode\n"));
	ret = U2F_Reg_Write(O_RDOMOD, 2);
	if (ret)
	{
	  ALTRO_err_get(ret, pid, code);
	  DEBUG_TEXT(P_ID_U2F, 5 ,("U2F_ReadOut: Error (%d) from U2F_Reg_Write: %s\n", ret, code));
	  return(RCC_ERROR_RETURN(0, ret)); 
	}
      }

      *osize = index;
      return(RCC_ERROR_RETURN(ret, U2F_LOW_MEM)); 
    }

    ret = ALTRO_Get(&in);
    if (ret) 
    {
      ALTRO_err_get(ret, pid, code);
      DEBUG_TEXT(P_ID_U2F, 5 ,("U2F_ReadOut: Error (%d) from ALTRO_Get: %s\n", ret, code));
      return(RCC_ERROR_RETURN(ret, U2F_ALTRO_FAIL)); 
    } 

    if (in.nbytes)
    { 
      DEBUG_TEXT(P_ID_U2F, 20 ,("U2F_ReadOut: %d bytes received\n", in.nbytes));
      /* copy the data from the altro_bulk_in_t structure to the data array */
      for(loop = 0; loop < in.nbytes; loop++) 
      { 
	data[index] = in.data[loop];
        if (index < 16)
          DEBUG_TEXT(P_ID_U2F, 30 ,("U2F_ReadOut: byte %d = 0x%02x\n", loop, in.data[loop]));
	index++;
      }
    }
    else
    {
      cont = 0;  
      DEBUG_TEXT(P_ID_U2F, 20 ,("U2F_ReadOut: Timeout, no data received. Terminating function\n"));
    }
  }
  
  if (mode & M_LAST)
  {
    DEBUG_TEXT(P_ID_U2F, 20 ,("U2F_ReadOut: Aborting read-out mode\n"));
    ret = U2F_Exec_Command(C_RDABORT, 0);
    if (ret)
    {
      ALTRO_err_get(ret, pid, code);
      DEBUG_TEXT(P_ID_U2F, 5 ,("U2F_ReadOut: Error (%d) from U2F_Exec_Command: %s\n", ret, code));
      return(RCC_ERROR_RETURN(0, ret)); 
    }
	    
    DEBUG_TEXT(P_ID_U2F, 20 ,("U2F_ReadOut: Disabling read-out mode\n"));
    ret = U2F_Reg_Write(O_RDOMOD, 0);
    if (ret)
    {
      ALTRO_err_get(ret, pid, code);
      DEBUG_TEXT(P_ID_U2F, 5 ,("U2F_ReadOut: Error (%d) from U2F_Reg_Write: %s\n", ret, code));
      return(RCC_ERROR_RETURN(0, ret)); 
    }
  }

  *osize = index;
  return(0);
}


/*************************/
u_int make_dummy_data(void) 
/*************************/
{
  return 48 + (u_int)(5 * (float)rand() / RAND_MAX);
}


/*******************************************************************/
void write_dummy_data(u_int which, u_int chip, u_int chan, u_int len) 
/*******************************************************************/
{
  typedef unsigned long long w40_t;
  static u_int data[512];
  u_int trailer = ((((w40_t)0x2aaa) << 26) + (((w40_t)len) << 16) + (((w40_t)0xa) << 12) + (((w40_t)chip) << 4) + (((w40_t)chan) << 4));
  u_int i, j = 0, n = len / 4, m = len % 4;

  for (i = 0; i < n; i++) 
  {
    data[2 * i + 0] = make_dummy_data(); 
    data[4 * i + 1] = make_dummy_data(); 
    data[4 * i + 2] = make_dummy_data(); 
    data[4 * i + 3] = make_dummy_data(); 
    j += 4;
  }
  
  for (i = 0; i < m; i++) 
  {
    data[2 * n + i] = make_dummy_data(); 
    j++;
  }
  
  for (; i < 4; i++) 
  {
    data[2 * n + i] = 0x2aa;
    j++;
  }
  
  data[2 * (n + 1) + 0] = (trailer >>  0) & 0xfffff;
  data[2 * (n + 1) + 1] = (trailer >> 20) & 0xfffff;
  j += 4;
  
  printf("Writing psuedo data to DMEM%d of size %d\n", which, j);
  U2F_DM_Write(which, j / 4, 0, data);
}
/***************************************************************/
unsigned int U2F_err_get(err_pack err, err_str pid, err_str code)
/***************************************************************/
{ 
  strcpy(pid, P_ID_ALTRO_STR);

  switch (RCC_ERROR_MINOR(err))
  {  
    case U2F_SUCCESS:    strcpy(code, U2F_SUCCESS_STR);     break;
    case U2F_FILE:       strcpy(code, U2F_FILE_STR);        break;
    case U2F_ALTRO_FAIL: strcpy(code, U2F_ALTRO_FAIL_STR);  break;
    case U2F_RANGE:      strcpy(code, U2F_RANGE_STR);       break;
    case U2F_ODD_BYTES:  strcpy(code, U2F_ODD_BYTES_STR);   break;
    case U2F_NODATA:     strcpy(code, U2F_NODATA_STR);      break;
    case U2F_INC_BYTES:  strcpy(code, U2F_INC_BYTES_STR);   break;
    case U2F_LOW_MEM:    strcpy(code, U2F_LOW_MEM_STR);     break;
    default:             strcpy(code, U2F_NO_CODE_STR);     return(RCC_ERROR_RETURN(0, U2F_NO_CODE)); break;
  }
  return(RCC_ERROR_RETURN(0, U2F_SUCCESS));
}
